/*
** Copyright 2001, Michael Pfeiffer	
*/

#include <KernelExport.h>
#include <Drivers.h>
#include <Errors.h>
#include <FindDirectory.h>
#include <image.h>
#include <stdio.h>
#include <string.h>
#include <graphic_driver.h>
#include <unistd.h>
#include "../include/AccelerantInterface.h"
#include "../include/DriverInterface.h"

#ifdef DEBUG
int log_fd = -1;
char *log_path = "/boot/var/log/wrapper.accelerant";
#endif

// area, so wrapper accelerant can retrieve the name of the wrapped accelerant 
static int                initialized = false;
       int                driver_fd = -1; /* referenced by hertz.c */
static struct SharedData *shared = NULL;
static area_id            shared_id = -1;
static image_id           accl_id = -1;

static void *   (*a_get_accelerant_hook)(uint32 feature, void *data) = NULL;
static status_t (*A_INIT_ACCELERANT)(int fd) = NULL;
static void     (*A_UNINIT_ACCELERANT)(void) = NULL;
       status_t (*A_SET_DISPLAY_MODE)(display_mode *mode_to_set) = NULL;
static status_t (*A_CLONE_ACCELERANT)(void *data);
       status_t (*A_GET_DISPLAY_MODE)(display_mode *mode_to_set) = NULL;

/* function prototypes */
static status_t init(uint32 feature, void* data, bool clone);

/*

The standard entry point.  Given a uint32 feature identifier, this routine
returns a pointer to the function that implements the feature.  Some features
require more information than just the identifier to select the proper
function.  The extra information (which is specific to the feature) is
pointed at by the void *data parameter.  By default, no extra information
is available.  Any extra information available to choose the function will be
noted on a case by case below.

*/
void * _EXPORT get_accelerant_hook(uint32 feature, void *data) {
#ifdef DEBUG
	char buffer[80];
	LOG("accelerant: get_accelerant_hook ");
	sprintf(buffer, "%d %x\n", (int)feature, (int)data);
	LOG(buffer);
#endif

	switch (feature) {
	case B_INIT_ACCELERANT:
		init(feature, data, false); return (void*)INIT_ACCELERANT;
		break;
	case B_CLONE_ACCELERANT:
		init(feature, data, true); return (void*)CLONE_ACCELERANT;
		break;
	case B_UNINIT_ACCELERANT:
		A_UNINIT_ACCELERANT = a_get_accelerant_hook(feature, data);
		return (void*)UNINIT_ACCELERANT;
		break;
	case B_SET_DISPLAY_MODE:
		A_SET_DISPLAY_MODE = a_get_accelerant_hook(feature, data);
		return (void*)SET_DISPLAY_MODE;
		break;
	case B_GET_DISPLAY_MODE:
		A_GET_DISPLAY_MODE = a_get_accelerant_hook(feature, data);
		return (void*)GET_DISPLAY_MODE;
		break;
	default:
		/* Delegate call to wrapped accelerant */
		if (a_get_accelerant_hook) return a_get_accelerant_hook(feature, data);
		LOG("accelerant: a_get_accelerant_hook is NULL\n");
		return 0;
	}
}


static bool load_accelerant(directory_which which, char* name) {
	char path[B_PATH_NAME_LENGTH+1];
	if (B_OK != find_directory(which, 0, false, path, B_PATH_NAME_LENGTH)) { accl_id = -1; return false; }
	strcat(path, "/accelerants/");
	strcat(path, name);
	LOG(path); LOG("\n");
	accl_id = load_add_on(path);
	return accl_id > 0;
}


static void unload_accelerant() {
	if (accl_id != -1) {
		unload_add_on(accl_id); accl_id = -1;
	}
}


static status_t get_shared_data() {
	area_id id = find_area(AREA_NAME);
	if (id == B_NAME_NOT_FOUND) return B_ERROR;
	shared_id = clone_area(AREA_NAME, (void**)&shared, B_ANY_ADDRESS, B_READ_AREA | B_WRITE_AREA, id);
	if (shared_id == B_BAD_VALUE || shared_id == B_NO_MEMORY || shared_id == B_ERROR) {
		shared_id = -1; return B_ERROR;
	}
	return B_OK;
}


static void release_shared_data() {
	if (shared_id != -1) {
		delete_area(shared_id); shared_id = -1; shared = NULL;
	}
}

static status_t init(uint32 feature, void* data, bool clone) {
	status_t err;
	if (initialized) return B_OK; // return, already initialized
	initialized = true;
	START_LOG;
	LOG("accelerant: enter INIT_ACCELERANT\n");
	// get accelerant name from shared area
	err = get_shared_data(); 
	// load accelerant
	if (err == B_OK && 
	    (load_accelerant(B_USER_ADDONS_DIRECTORY, shared->accelerant) ||
	   	load_accelerant(B_BEOS_ADDONS_DIRECTORY, shared->accelerant))) {
	   	
	   	err = get_image_symbol(accl_id, "get_accelerant_hook", B_SYMBOL_TYPE_TEXT, (void**)&a_get_accelerant_hook);

	    if (err != B_OK) {
	    	a_get_accelerant_hook = NULL; 
		} else {
			if (clone) {
				A_CLONE_ACCELERANT = a_get_accelerant_hook(B_CLONE_ACCELERANT, data);
			} else {
				A_INIT_ACCELERANT = a_get_accelerant_hook(B_INIT_ACCELERANT, data);
			}
		}
	} else {
		err = B_ERROR;
	}
	
	if (err != B_OK) {
		unload_accelerant();
		release_shared_data();
		STOP_LOG;
	} else {
		LOG("leave INIT_ACCELERANT\n");
	}
	
	return err;
}


status_t INIT_ACCELERANT(int fd) {
	driver_fd = fd;
	if (A_INIT_ACCELERANT) {
		status_t err = A_INIT_ACCELERANT(fd);
		if (err == B_OK) {
			load_tv_out_add_on(fd);
		}
		return err;
	} 
	return B_ERROR;
}


status_t CLONE_ACCELERANT(void* data) {
	LOG("accelerant: enter CLONE_ACCELERANT\n");
	if (A_CLONE_ACCELERANT) return A_CLONE_ACCELERANT(data);
	return B_ERROR;
}


void UNINIT_ACCELERANT(void) {
	LOG("accelerant: enter UNINIT_ACCELERANT\n");
	unload_tv_out_add_on();
	if (A_UNINIT_ACCELERANT) A_UNINIT_ACCELERANT();
	unload_accelerant();
	release_shared_data();
	STOP_LOG;
}


void print_mode(display_mode *m) {
#ifdef DEBUG
	char buffer[256];
	LOG("display_mode:\n");
	#define P(f) sprintf(buffer, "  " #f " = %d\n", (int)m->f); LOG(buffer);
	P(space);
	P(virtual_width);
	P(virtual_height);
	P(h_display_start);
	P(v_display_start);
	P(flags);
	P(timing.pixel_clock);
	P(timing.h_display);
	P(timing.h_sync_start);
	P(timing.h_sync_end);
	P(timing.h_total);
	P(timing.v_display);
	P(timing.v_sync_start);
	P(timing.v_sync_end);
	P(timing.v_total);
	P(timing.flags);
#endif
}

status_t SET_DISPLAY_MODE(display_mode *mode_to_set) {
	status_t err;
	
	LOG("accelerant: enter SET_DISPLAY_MODE\n");
	print_mode(mode_to_set);
	// ASSERT(A_SET_DISPLAY_MODE != NULL)
	
	err = tv_out_set_display_mode(mode_to_set, A_SET_DISPLAY_MODE);
	if (err == B_OK) return B_OK;
	
	err = hack_display_mode(mode_to_set);
	if (err == B_OK) return B_OK;
	
	// delegate call to wrapped accelerant
	err = A_SET_DISPLAY_MODE(mode_to_set);
#ifdef DEBUG
	print_mode(mode_to_set);
	{ 
		char buffer[80];
		sprintf(buffer, "err = %d 0x%x\n", (int) err, (int) err);
		LOG(buffer);
	}
#endif
	return err;
}

status_t GET_DISPLAY_MODE(display_mode *mode) {
	status_t err;
	
	LOG("accelerant: enter GET_DISPLAY_MODE\n");
		
	err = tv_out_get_display_mode(mode);
	if (err == B_OK) return B_OK;
	
	// delegate call to wrapped accelerant
	err = A_GET_DISPLAY_MODE(mode);
#ifdef DEBUG
	print_mode(mode);
	{ 
		char buffer[80];
		sprintf(buffer, "err = %d 0x%x\n", (int) err, (int) err);
		LOG(buffer);
	}
#endif
	return err;
}


